#include <string>
#include <event.h> /* Using libevent for the polling facilities */
#include "autoptr.hh"

struct ConnHandler: public ptrable
{
    ConnHandler() { }
    virtual ~ConnHandler() { }
    
    virtual void Init(struct Conn* ) { }
    
    /* 0=not done, 1=done after writebuffer empties, 2=done immediately */
    enum OverType { NotOver, OverAfterWrite, OverNow };
    
    virtual OverType IsDone() const { return OverAfterWrite; }
    
    virtual void GotTimeout(struct Conn* ) { }
    
    virtual void GotInput(struct Conn* c, unsigned old_data_length) = 0;
};

/******************/
/* manipulation of events (wrapped into Conn objects) */
/******************/
struct Conn
{
    int fd;
    event ev;
    std::string Received;
    std::string Sending;
    short flags;
    double timeout;
    
    /* Handler is needed here so that UpdateEvent() can
     * update the event. It is not used for calling directly. */
    void (*handler)(int,short,void*);
    
    autoptr<ConnHandler> obj;
    
    Conn() : fd(0),ev(),Received(),Sending(),flags(0),timeout(0),handler(),obj() { }
};

static Conn* _SetupEvent(Conn* c, int fd,short flags,
    void(*handler)(int,short,void*),
    double timeout = 0)
{
    struct timeval tv = { (int)timeout, (int)(timeout * 1e6) % 1000000 };

    event_set(&c->ev, fd, flags, handler, (void*)c);
    event_add(&c->ev, timeout != 0 ? &tv : 0);
    c->fd = fd;
    c->flags = flags;
    c->handler = handler;
    c->timeout = timeout;
    return c;
}
static Conn* UpdateEvent(Conn* c, short new_flags, double new_timeout = 0.0)
{
    if(new_timeout == 0.0 && c->timeout == 0.0
    && new_flags == c->flags) return c;
    
    /* TODO: Figure out how to make this issue EPOLL_CTL_MOD
     * instead of EPOLL_CTL_DEL + EPOLL_CTL_ADD
     */
    event_del(&c->ev);
    return _SetupEvent(c, c->fd, new_flags, c->handler, new_timeout);
}
static Conn* CreateEvent(int fd, short flags, void (*handler)(int,short,void*),
    ConnHandler* obj = 0,
    double timeout = 0)
{
    Conn* c = new Conn;
    c->obj = obj;
    _SetupEvent(c, fd,flags,handler, timeout);
    if(c->obj) c->obj->Init(c);
    return c;
}
static void EndEvent(Conn* c)  
{
    event_del(&c->ev);
    close(c->fd);
    delete c;
}
/******************/
